/*
vim:ts=2:sw=2:
RJVB: cleaned up the tab mess...
 * commands.c -- POP3 command handlers
 * hotwayd is a POP3-HTTPMail gateway. It will allow you to
 * check your HOTMAIL account from your unix box.
 *
 * Created by: Espeleta Tomas <espeleta@libero.it>, 12-Apr-2001
 * POP3 code based on nupop-0.4, by Kevin Stone <kstone@nuvox.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "libghttp-1.0.9-mod/ghttp.h"
#include "libghttp-1.0.9-mod/ghttp_constants.h"
#include "httpmail.h"
#include "hotwayd.h"
#include "inet.h"
#include <sys/time.h>

/* Global Vars */
KEYPAIR **cookies = NULL;
FOLDER_STRUCT *inbox_props = NULL;
ghttp_request *request = NULL;
time_t time_getlist; /* time when getlist() was last called, used to update LIST */
#define GETLIST_TIMEOUT 120 /* 2 minute timeout for list command to refresh */
const char *badpwdtxt[] = { "Bad Password", "Invalid login." };
const int badpwdcnt = sizeof(badpwdtxt) / sizeof(char *);

int lazy_dump(char *pos, int len, int *curpos_p, int *curline_p, int maxlines);
static void prepare_and_send(ghttp_request *request, char *url);
static int set_authinfo(ghttp_request *request, char *UserName, char *Password,
			char *WWWAuthenticate, int maintainlogged);
static int clear_authinfo(ghttp_request *request);


extern int cur_message;
extern int top_lines;  /* how many lines to display using top */
extern int pastheader; /* used for displaying the message using top */
char buffer[N_BUFFER];
char *serverurl=NULL;
char xml_length[N_XML_LENGTH];
char *UserName = NULL;  /* allocate space depending on size later */
char PassWord[N_PASSWORD];
char *WWWAuthenticate;
static int unauth_cnt = 0; /* check for unauth loop on proxy */


/* proxy variables are setup via command line arguments */
char *proxy=NULL;  /* proxy address:port */
char *proxy_username=NULL;
char *proxy_password=NULL;
extern char *folder;

/* Extern stuff */
extern char *dir_names_request; /* request for folder hrefs */
extern char *dir_props_request; /* request for folder properties */
extern char *msg_props_request; /* request for messages within folder props */
int ParseXMLMsgList(ghttp_request *request, FOLDER_STRUCT *folder_props);
int ParseXMLFolderProps(ghttp_request *request, char *foldername, FOLDER_STRUCT *folder_props);
char *ParseXMLFolderRoot(ghttp_request *request, FOLDER_STRUCT *folder_props);


/* general counters/flags */
int oldstatus      = 0;
int cookiecount    = 0;
int auth_required  = 0;
int listdownloaded = 0; /* got the inbox list yet? */
extern int log_level;

typedef struct HTTP_Servers{
	char *domain;
	char *URL;
} HTTP_Servers;

const HTTP_Servers default_server[]= {
	{ "hotmail.com", "http://oe.hotmail.com/cgi-bin/hmdata" },
	{ "msn.com", "http://oe.msn.msnmail.hotmail.com/cgi-bin/hmdata" },
	{ "lycos.co.uk", "http://webdav.lycos.co.uk/httpmail.asp" },
	{ "lycos.ch", "http://webdav.lycos.de/httpmail.asp" },
	{ "lycos.de", "http://webdav.lycos.de/httpmail.asp" },
	{ "lycos.es", "http://webdav.lycos.es/httpmail.asp" },
	{ "lycos.it", "http://webdav.lycos.it/httpmail.asp" },
	{ "lycos.at", "http://webdav.lycos.at/httpmail.asp" },
	{ "lycos.nl", "http://webdav.lycos.nl/httpmail.asp" },
	{ "spray.se", "http://webdav.spray.se/httpmail.asp" },
};


/* Defintions for the types of http authorisation methods */
#define UNKNOWN 0
#define BASIC 1
#define DIGEST 2

typedef struct Auth_Method {
	char *wwwauthstr;
	int auth_type;
} Auth_Method;

const Auth_Method auth_method[] = {
  { "Basic", BASIC },
  { "Digest", DIGEST },
};

int default_servers= sizeof(default_server)/sizeof(HTTP_Servers);
int auth_methods   = sizeof(auth_method)/sizeof(Auth_Method);

/* This is the authorisation type we recieved from the server */
int auth_name = 0;

/* User Agent: was hotwayd, but M$ maybe does'nt like it */
const char *myAgent = "Outlook-Express/5.5 (MSIE 5.0; Windows 98; TmstmpExt)";

/* Additional Header, not defined in ghttp.h */
const char *http_hdr_Brief				= "Brief";
const char *http_hdr_Translate		= "Translate";
const char *http_hdr_Allow_Rename = "Allow-Rename";

/* ---------------------------------------------------------------------
 *	Here begin internal functions that handle message-popping
 *	from HTTPMail server
 * ---------------------------------------------------------------------
 */

/* ---------------------------------------------------------------------
 * SET Cookies, taking them from array of structures and verifing domain
 */
void my_set_cookies(ghttp_request *request, char *url)
{
	int i;
	char mybuf[N_MYBUF];

	buffer[0] = '\0';

	for (i = 0; i < cookiecount; i++)		/* control if exists */
			/* 20020923: need to test for cookies[i]!=NULL (in case of manual retry-to-login?!) */
		if( cookies[i] && cookies[i]->domain && strstr(url, cookies[i]->domain) != NULL) {
			snprintf(mybuf, N_MYBUF, "%s=%s; ", cookies[i]->key, cookies[i]->value);
			strlcat(buffer, mybuf, N_BUFFER);
		}

	i = strlen(buffer);
	if (i > 0) {
		buffer[i-1] = '\0'; /* truncate last space */
		ghttp_set_header(request, "Cookie", buffer);
	}
}
/* ---------------------------------------------------------------------
 * Parse and save all cookies passed by response
 */
void save_cookies(ghttp_request *request)
{
	int i, j = -1, n;
	char *k, *key, *value, *cookiedomain = NULL;
	char *tok, *which, pair[N_PAIR];
	char **hdrs_list;

	/* We try to find Set-Cookie headers... */

	if ((hdrs_list = ghttp_get_header_values(request, http_hdr_Set_Cookie)) != NULL) {
		for (n = 0; hdrs_list[n] != NULL; n++) {
			j=-1;
			strlcpy(buffer, hdrs_list[n],N_BUFFER);

			/* Explode key/value; domain; path */
			for (which = buffer; (tok = strtok(which, ";")) != NULL; which = NULL) {

	/* Now tok contains a pair <key="value">	*/
	strlcpy(pair, tok, N_PAIR);

	if ((k = index(pair, '=')) != NULL) {
		*k = '\0';
		value = k + 1;

		for (key = pair; *key == ' '; ) key++;
		if (value[strlen(value)-1] == '\"') value[strlen(value)-1] = '\0';
		if (value[0] == '\"') value++;

		if (!strcmp(key, "domain")) { /* Read domain for this cookie */
			cookiedomain = strdup(value);
			if (j != -1) cookies[j]->domain = cookiedomain;
		}
		else if (strcmp(key, "path")) { /* Read key/value for this cookie, ignoring path*/

			for (i = 0; i < cookiecount; i++)		/* control if exists */
				if( cookies[i] && !strcmp(key, cookies[i]->key)){
			j = i;
		 }

			if (j == -1) {		/* it does not exist: we stack it, else overwrite */
				j = cookiecount;
				cookies[j] = (KEYPAIR*) calloc (1, sizeof(KEYPAIR));
				cookiecount++;
			}
			cookies[j]->key = strdup(key);	/* let's store it */
			cookies[j]->value = strdup(value);
			cookies[j]->domain = cookiedomain;
		}
	}
			}
		}
	}
}
/* ---------------------------------------------------------------------
 * Propfind makes a PROPFIND call on myurl 
 */
int propfind(ghttp_request *request, char *myurl, int whichrequest)
{
	int status;
	
	/* Settings of our connection	 */
	ghttp_clean(request);
	ghttp_set_uri(request, myurl);

	if (ghttp_set_type(request, ghttp_type_propfind) == -1)
		return -1;

	/* Set Body request & Depth header */
	switch(whichrequest) {
	case MSGFOLDERROOT:
	  if (ghttp_set_body(request, dir_names_request, (int)(strlen(dir_names_request)) ) == -1)
		return -1;
	  ghttp_set_header(request, http_hdr_Depth, "0");
	  snprintf(xml_length, N_XML_LENGTH, "%d", (int)(strlen(dir_names_request)));
	  break;
	case INBOXPROPS:
	  if (ghttp_set_body(request, dir_props_request, (int)(strlen(dir_props_request)) ) == -1)
		return -1;
	  ghttp_set_header(request, http_hdr_Depth, "1, noroot");
	  snprintf(xml_length, N_XML_LENGTH, "%d", (int)(strlen(dir_props_request)));
	  break;
	case MSGLIST:
	  if (ghttp_set_body(request, msg_props_request, (int)(strlen(msg_props_request)) ) == -1)
		return -1;
	  ghttp_set_header(request, http_hdr_Depth, "1, noroot");
	  snprintf(xml_length, N_XML_LENGTH, "%d", (int)(strlen(msg_props_request)));
	  break;
	default:
	  if (log_level > 0) 
	    LOG("Internal error: Received invalid request to propfind in %s",__FILE__);
	  httpmail_destroy();
	  exit(-1);
	}
	  
	/* Set Headers */
	ghttp_set_header(request, http_hdr_Content_Type, "text/xml");
	ghttp_set_header(request, http_hdr_Brief, "t");
	ghttp_set_header(request, http_hdr_Content_Length, xml_length);
	/* clear_authinfo(request); */
	if (auth_required) {
	  /* First we have to find out what type of auth the server requested */
	  int i;
	  for (i=0; i<auth_methods && !auth_name; i++) {
	    if (!strncasecmp(auth_method[i].wwwauthstr, WWWAuthenticate, strlen(auth_method[i].wwwauthstr))) {
	      auth_name=auth_method[i].auth_type;
	      break;
	    }
	  }
	  set_authinfo(request, UserName, PassWord, WWWAuthenticate, 0);

	}
	if (whichrequest == MSGLIST)
	  set_authinfo(request, UserName, PassWord, NULL, 1);

	/* now prepare and send our http request that we've built */
	prepare_and_send(request, myurl);
	
	/* Save response cookies */
	save_cookies(request);
				
	if ((status = ghttp_status_code(request)) == AUTHORIZATION) {
	  int i;
	  if (unauth_cnt++ > 3) /* make it 4 times just to be sure :-) */
	    return PASSWORDFAILED;
	  /* check for password failure via ghttp_reason_phrase, this is the only
	   * way I can see to distinguish the various AUTHORIZATION states. Doing
	   * it this way prevents invalid password being returned when the
	   * response was actually unauthorized */
	  for (i = 0; i < badpwdcnt; i++) {
	    if (!strcmp(badpwdtxt[i], ghttp_reason_phrase(request))) 
		return PASSWORDFAILED;
	  }
	} else
	  auth_required = 0;

	return status;
}

/* ---------------------------------------------------------------------
 * Handles replies: redirections and authorizations are recognized!
 * returns reply on break, else returns 0;
 */
int handlereplies(ghttp_request *request, int reply)
{
	const char *wwwauth;

	if (reply != AUTHORIZATION)
	  unauth_cnt=0; /* reset the unauthorised counter */
	
#ifdef DEBUG
	fprintf( stderr, "handlereplies(%d): ", reply ) ; fflush( stderr );
#endif
	if (reply == AUTHORIZATION) {
		if ((wwwauth = ghttp_get_header(request, http_hdr_WWW_Authenticate)) != NULL) {
			if (WWWAuthenticate)
			  free(WWWAuthenticate);
			WWWAuthenticate = strdup(wwwauth);
			auth_required = 1;
#ifdef DEBUG
	 fprintf( stderr, "reply=AUTHORIZATION; %s\n", wwwauth );
#endif
		}
		return 0;
	}
	else if (reply == MULTISTATUS || reply == PASSWORDFAILED || reply == INVALIDUSER) {
#ifdef DEBUG
		fprintf( stderr, "reply=%d\n", reply );
#endif
		return reply;
	}
	else if (reply == OKSTATUS || reply == MOVED) {
#ifdef DEBUG
	  fprintf( stderr, "reply=%d\n", reply );
#endif
	  /* need to reset the authorization info if we move to a new server */
	  clear_authinfo(request); 
	  return 0; /* Sometimes we get 200. We will ignore it. */
	}
	else {
#ifdef DEBUG
		fprintf( stderr, "ERROR reply=%d\n", reply );
#endif
		return ERROR; /* FIXME: this isn't getting passed thu */
	}
}


/* use this function to fill in the URL if we get moved response */
static char *get_moved_url(ghttp_request *req) {
  char *new_url=NULL;
  
  if (ghttp_get_header(request, http_hdr_Location) != NULL) {
	new_url=strdup(ghttp_get_header(request, http_hdr_Location));
#ifdef DEBUG
	fprintf( stderr, "reply=MOVED; %s\n", new_url );
#endif
  }

  return new_url;
}


static int set_authinfo(ghttp_request *request, char *UserName, char *Password,
			char *WWWAuthenticate, int maintainlogged) {
  int ret;
  clear_authinfo(request);
  switch(auth_name) {
  case DIGEST:
    ret=ghttp_set_digestinfo(request, UserName, PassWord,
			     WWWAuthenticate, maintainlogged);
    break;
  case BASIC:
    ret=ghttp_set_authinfo(request, UserName, PassWord);
    break;
  case UNKNOWN:
  default:
    /* opps couldn't figure out auth... */
    if (log_level > 0) 
      LOG("%s:%d - couldn't figure out auth type from server request",__FILE__,__LINE__);
    break;
  }
  
  return ret;    
}

static int clear_authinfo(ghttp_request *request) {
  int ret;
  ret=ghttp_set_digestinfo(request, NULL, NULL, NULL, 0);
  ret=ghttp_set_authinfo(request, NULL, NULL);
  return ret;
}

/* ---------------------------------------------------------------------
 * Login, try to connect and login to the hotmail server
 * returns 1 on success and 0 on failure
 */
int dologin(ghttp_request *request) 
{
    int reply = 0;
    char *root_href = NULL;

    /* A cycle of propfind requests, until MULTISTATUS is returned */
    while ((reply = propfind(request, serverurl, MSGFOLDERROOT)) != MULTISTATUS) {
      if (reply == MOVED) {
	free(serverurl);
	serverurl = get_moved_url(request);
      }
      if (handlereplies(request, reply)) 
	break;
    }
    /* So here we have either logged in or bombed out */
    unauth_cnt=0; /* reset unauth counter */

    /* lets try to find the root href of the mailbox so we can call dir_props */
    if (reply == MULTISTATUS &&	(root_href = ParseXMLFolderRoot(request, inbox_props)) != NULL) {
      if (serverurl) free(serverurl);
      serverurl = strdup(root_href);  /* copy folder root location */
      if (root_href) free(root_href);

      /* So at this point we are logged into hotmail, but now let's see if we
       * can get into the folder that we have requested */
      /* A cycle of propfind requests, until MULTISTATUS is returned */
      while ((reply = propfind(request, serverurl, INBOXPROPS)) != MULTISTATUS) {
	if (reply == MOVED) {
	  free(serverurl);
	  serverurl = get_moved_url(request);
	}
        if (handlereplies(request, reply))
	  break;
      }
      unauth_cnt=0; /* reset unauth counter */

      /* Lets parse the XML and see if we can find the folder we were after */
      return ParseXMLFolderProps(request, folder, inbox_props);
    }

    /* otherwise return our bombed out reply */
    return reply;
}

/* ---------------------------------------------------------------------
 * Get List of messages from server
 */
int getlist(ghttp_request *request)
{
	int reply = 0, downloaded = 0;

	time(&time_getlist);
	
	/* A cycle of propfind requests, until MULTISTATUS is returned */
	while ((reply = propfind(request, inbox_props->href, MSGLIST)) != MULTISTATUS) {
	  if (reply == MOVED) {
		free(inbox_props->href);
		inbox_props->href = get_moved_url(request);
	  }
	  if (handlereplies(request, reply))
		break;
	}
	if (reply == MULTISTATUS && ParseXMLMsgList(request, inbox_props))
		downloaded = 1;
	unauth_cnt=0; /* reset unauth counter */
	return downloaded;
}

/* ---------------------------------------------------------------------
 * Get messages from server
 */
char *getmessage(ghttp_request *request, int num_msg)
{

	if (inbox_props->list_props ){
		if( inbox_props->list_props[num_msg]->message ){
		  int curpos = 0;
		  int curline = 0;
		  PSOUT("+OK We have that cached\r\n");
		  lazy_dump(inbox_props->list_props[num_msg]->message, -1, &curpos, &curline, top_lines);
		  PSOUT(".\r\n");
		  return( inbox_props->list_props[num_msg]->message );
		}
		else if( inbox_props->list_props[num_msg]->href ) {
			/* Settings of our connection	 */
			ghttp_clean(request);
			ghttp_set_uri(request, inbox_props->list_props[num_msg]->href);
			ghttp_set_type(request, ghttp_type_lazy);

			/* Set Headers */
			ghttp_set_header(request, http_hdr_Accept, "message/rfc822, */*");
			ghttp_set_header(request, http_hdr_Translate, "f");
			set_authinfo(request, UserName, PassWord, NULL, 1);

			/* Prepare and process the connection */
			prepare_and_send(request, inbox_props->msg_props[num_msg]->href);

			cur_message = num_msg;

			return NULL;
		}
	}
	return NULL;
}

/* ---------------------------------------------------------------------
 * Delete messages from server (moving it to hm trash)
 */
int delmessage(ghttp_request *request, int num_msg)
{
	char *msg_name;

	if (!inbox_props->trash)
		return 0;

	if (inbox_props->list_props && inbox_props->list_props[num_msg]->href ) {
		/* Set message destination (trash) */
		msg_name = strrchr(inbox_props->list_props[num_msg]->href, '/');
		msg_name++;
		strlcpy(buffer, inbox_props->trash, N_BUFFER);
		strlcat(buffer, msg_name, N_BUFFER);
			
		/* Settings of our connection	 */
		ghttp_clean(request);
		ghttp_set_uri(request, inbox_props->list_props[num_msg]->href);
		ghttp_set_type(request, ghttp_type_move);

		/* Set Headers */
		ghttp_set_header(request, http_hdr_Destination, buffer);
		ghttp_set_header(request, http_hdr_Allow_Rename, "t");
		ghttp_set_header(request, http_hdr_Content_Length, "0");
		set_authinfo(request, UserName, PassWord, NULL, 1);

		prepare_and_send(request, inbox_props->list_props[num_msg]->href);
		
		/* is response ok? (should be 201) */
		if (ghttp_status_code(request) >= 200 && ghttp_status_code(request) < 300)
		  return 1;
		else
		  return 0;
	}
	return 0;
}

/* ---------------------------------------------------------------------
 * Functions called by hotwayd.c. They are a translation POP3 -> HTTPMail
 * ---------------------------------------------------------------------
 */

long httpmail_nemails(void)
{
	return inbox_props->listcount;
}

void httpmail_destroy(void)
{
	int i;

	/* Free memory */
	if (inbox_props != NULL) {
		free(inbox_props->href);
		free(inbox_props->trash);
		free(inbox_props->name);
		if (inbox_props->visiblecount > 0 && listdownloaded) {
			for (i = 0; i < inbox_props->visiblecount; i++) {
				if( inbox_props->msg_props[i]->href ){
					free(inbox_props->msg_props[i]->href);
				}
				if( inbox_props->msg_props[i]->message ){
					free(inbox_props->msg_props[i]->message);
				}
				free(inbox_props->msg_props[i]);
			}
		}
		listdownloaded=0; /* 20020924: reset listdownloaded variable */
		free(inbox_props->msg_props);
		inbox_props->msg_props= NULL;
		free(inbox_props->unread_props);
		inbox_props->unread_props=NULL;
		free(inbox_props);
	}
	inbox_props= NULL;

	/* ensure that we reset cookiecount back to 0! */
	while(cookiecount>0) {
		cookiecount--;
		if( cookies[cookiecount] ){
			free(cookies[cookiecount]->key);
			free(cookies[cookiecount]->value);
			free(cookies[cookiecount]->domain);
			free(cookies[cookiecount]);
			cookies[cookiecount]= NULL;
		}
	}
	
	if (UserName)  free(UserName);
	UserName=NULL;
	if (serverurl) free(serverurl);
	serverurl=NULL;
	if (cookies)   free(cookies);
	cookies= NULL;
	if (request)   ghttp_request_destroy(request);

}

/* ---------------------------------------------------------------------
 * Setup the data structures and global variables then call dologin
 * Return 1 on successful login and 0 on failure (via dologin)
 */
int httpmail_authenticate_user(char *user, char *pass)
{
	/* Allocate a new empty request object */
	request = ghttp_request_new();

	/* Allocate cookies structure and inbox props if not already allocated */
	if (!cookies) cookies = (KEYPAIR **) calloc (MAX_COOKIES, sizeof(KEYPAIR *));
	if (!inbox_props) inbox_props = (FOLDER_STRUCT *) calloc(1, sizeof(FOLDER_STRUCT));

	if (UserName) free(UserName); /* get rid of any stale username */
	UserName = (char *) malloc( (strlen(user) + 1) * sizeof(char) );
	
	/* Set username/password for authorization-header */	
	strlcpy(UserName, user, (strlen(user) + 1));
	strlcpy(PassWord, pass, N_PASSWORD);

	return dologin(request);
}

/* this function checks we have a user name of the form xxx@hotmail.com
 * or xxx@msn.com or xxx%hotmail.com or xxx%msn.com */
int validate_username(char *user)
{
  char *domain= index( user, '@');
  char *folder_pos= index( user, '/');
  if (serverurl) free(serverurl);
  serverurl=NULL;

  if (!domain) domain=index(user,'%'); /* check for % sign as well */
  if (!domain) {
	    return E_LOGIN_NO_DOMAIN; /* neither an @ nor % was there */
  } else {
		int i;
		*domain='@'; /* make sure the @ sign is there, overwrite any % */
		domain+= 1;
		if (folder_pos) *folder_pos='\0'; /* ignore the folder section for the moment */
		for( i= 0; i< default_servers && !serverurl; i++ ){
		  /* 20030128 RJVB: do case-insensitive comparison of the servername...! */
		  if( strcasecmp( default_server[i].domain, domain)== 0 ){
			serverurl=strdup(default_server[i].URL);
		  }
		}
		if (folder_pos) *folder_pos='/'; /* put folder back */
  }
	
  if( !serverurl ){
	    return E_LOGIN_UNRECOGNISED_DOMAIN;
  }
  return E_LOGIN_OK; /* username validates */
}


void httpmail_stat(cmdtable *command)
{
	long maildrop_size	 = 0,
		n_not_deleted_emails = 0,
		i;

	/* GET MESSAGES LIST */
	if (inbox_props->href) {
		getlist(request);
		listdownloaded = 1;
		for (i = 0; i < inbox_props->listcount; i++) {
			  /* 20030130 RJVB: FIXME: deleted mails seem to show up as empty messages, at the end of the list
				 \ can we just not increment n_not_deleted_emails when length==0, or only when there is no more
				 \ non-zero-length message *after* the current one?
				 \ But in fact, visiblecount should not have included empty
				 messages! */
		      /* 20030421 DAVE: fixed visiblecount inside xmlstuff.c. stat
				 \ and list run update_visiblecount before filling the message
				 \structs now. Added a check if running in debug mode also.
				 */
			n_not_deleted_emails++;
			maildrop_size += inbox_props->list_props[i]->length;
#if DEBUG
			if (!inbox_props->msg_props[i]->length)
			  fprintf(stderr, "warning: msg %lu has length 0, that is strange\n", i);
#endif

		}
	}

	snprintf(buffer, N_BUFFER, "+OK %lu %lu", n_not_deleted_emails, maildrop_size);
	register_command( command, buffer, 0, NULL, NULL );
	PSOUT(buffer);
	PCRLF;
}

void httpmail_list(cmdtable *command, char *arg)
{
	long i;
	short cache_flag=1;

	if ( ( (!listdownloaded || (time(NULL)-time_getlist > GETLIST_TIMEOUT) ) && inbox_props->href) ) {
		getlist(request);
		listdownloaded = 1;
		cache_flag--;
	}

	if (arg == NULL) {
	    if (!cache_flag) {
		    PSOUT("+OK Mailbox scan listing follows");
		} else {
		    PSOUT("+OK Cached mailbox scan listing follows");
		}
		PCRLF;
 		for (i = 0; i < inbox_props->listcount; i++) {
 			snprintf(buffer, N_BUFFER, "%lu %u\015\012", (i + 1), inbox_props->list_props[i]->length);
			PSOUT(buffer);
		}
		PSOUT(".");
		PCRLF;
		return;
	}
 	if ((i = strtoul(arg, NULL, 10)) && (i <= inbox_props->listcount) && (i > 0)) {
 		snprintf(buffer, N_BUFFER, "+OK %lu %u", i, inbox_props->list_props[i - 1]->length);
		register_command( command, buffer, 1, arg, NULL );
		PSOUT(buffer);
		PCRLF; 
	} else {
		register_command( NULL, NULL, 0, NULL, NULL );
		PSOUT("-ERR No such message");
		PCRLF;
	}
}

void httpmail_rset(cmdtable *command)
{
	register_command( NULL, NULL, 0, NULL, NULL );
	PSOUT("+OK Mailbox has been reset");
	PCRLF;
}

void httpmail_delete(cmdtable *command, char *arg)
{
	long i;

	register_command( NULL, NULL, 0, NULL, NULL );	
 	if ((i = strtoul(arg, NULL, 10)) && (i <= inbox_props->listcount) && (i > 0)) {
		if (delmessage(request, i-1)) {
			PSOUT("+OK Message deleted");
			PCRLF;
		}
		else {
			PSOUT("-ERR Error deleting message");
			PCRLF;
		}
	}
	else {
		PSOUT("-ERR No such message");
		PCRLF;
	}
}

char *gen_uidl(char *dest, char *source)
{
	int i, j;

	if( !source ){
		strcpy( dest, "<DELETED>" );
	}
	else{
		for (j = 0, i = strlen(source) - 1; i >= 0 && j < 31; i--, j++)
			dest[j] = (source[i] != '/') ? source[i] : 'Z';
		dest[j] = '\0';
	}
	return dest;
}

void httpmail_uidl(cmdtable *command, char *arg)
{
	char buf[32];
	long i;

	if (!listdownloaded && inbox_props->href) {
		getlist(request);
		listdownloaded = 1;
	}

	if (arg == NULL) {
	  char *response= (char*) calloc(inbox_props->listcount* N_BUFFER+2, sizeof(char));
	  snprintf(buffer, N_BUFFER, "+OK unique-id listing follows\015\012");
    if (response) strcat(response, buffer);
	  PSOUT(buffer);
 		for (i = 0; i < inbox_props->listcount; i++) {
 			snprintf(buffer, N_BUFFER, "%lu %s\015\012", (i + 1), gen_uidl(buf, inbox_props->list_props[i]->href));
			if( response ){
				strcat( response, buffer );
			}
			PSOUT(buffer);
		}
		PSOUT(".");
		if( response ){
			strcat( response, "." );
		}
		PCRLF;
		register_command( command, response, 0, NULL, NULL );
		if( response ){
			free(response);
		}
		return;
	}
 	else if ((i = strtoul(arg, NULL, 10)) && (i <= inbox_props->listcount) && (i > 0)) {
 		snprintf(buffer, N_BUFFER, "+OK %lu %s\015\012", i, gen_uidl(buf, inbox_props->list_props[i - 1]->href));
		PSOUT(buffer);
	/* Don't need the PCRLF here as buffer already has CR/LF added */
	} else {
		register_command( NULL, NULL, 0, NULL, NULL );
		PSOUT("-ERR No such message");
		PCRLF;
	}
}

char *next_eol( const char *line, int *len )
{ char *l= NULL;
	*len= 0;
	if( line && *line ){
		if( (l= strstr(line, "\n")) ){
			*len= 1;
		}
		else if( (l= strstr(line, "\r\n")) ){
			*len= 2;
		}
		else if( (l= strstr(line, "\r")) ){
			*len= 1;
		}
	}
	return(l);
}

char *next_emptyline( const char *line, int *len )
{ char *l=NULL;
	*len= 0;
	if( line && *line ){
		if( (l= strstr(line, "\n\n")) ){
			*len= 2;
		}
		else if( (l= strstr(line, "\r\n\r\n")) ){
			*len= 4;
		}
		else if( (l= strstr(line, "\r\r")) ){
			*len= 2;
		}
	}
	return(l);
}

char *next_dotline( const char *line, int *len )
{ char *l= NULL;
	*len= 0;
	if( line && *line ){
		if( (l= strstr(line, "\n.")) ){
			*len= 2;
		}
		else if( (l= strstr(line, "\r\n.")) ){
			*len= 3;
		}
		else if( (l= strstr(line, "\r.")) ){
			*len= 2;
		}
	}
	return(l);
}

char *concat( char *text, char *addition )
{
	if( !addition ){
		return( text );
	}
	if( text ){
		if( (text= (char*) realloc( text, (strlen(text) + strlen(addition)+ 1) * sizeof(char) )) ){
			strcat( text, addition );
		}
	}
	else{
		text= strdup(addition);
	}
	return(text);
}

/* 20020927 RJVB: lowlevel message sending routine. Takes care of byte-stuffing, i.e.
 \ prepending a period to lines that start with a period.
 \ 20030128 RJVB: returns the output message.
 */
char *send_message( char *msg, char *response )
{ char *ldot= NULL, D= 0;
  int len;
	if( msg ){
		ldot= next_dotline(msg, &len);
		if( msg[0]== '.' || ldot ){
			while( msg && (*msg== '.' || ldot) ){
				if( ldot ){
					ldot+= len-1;
					D= *ldot;
					*ldot= '\0';
				}
				if( *msg== '.' ){
					PSOUT( "." );
					if( response ){
						response= concat( response, "." );
					}
				}
				PSOUT( msg );
				if( response ){
					response= concat( response, msg );
				}
				if( ldot ){
					*ldot= D;
					msg= ldot;
					ldot= next_dotline(msg, &len);
				}
				else{
					msg= NULL;
				}
			}
			  /* an additional newline seems to be required here to be compatible with other servers: */
			PCRLF;
			if( response ){
				response= concat( response, "\015\012" );
			}
		}
		else{
			PSOUT( msg );
			if( response ){
				response= concat( response, msg );
			}
		}
	}
	return(response);
}

void httpmail_retrieve(cmdtable *command, char *arg)
{
  long i;

  if ((i = strtoul(arg, NULL, 10)) && (i <= inbox_props->listcount) && (i > 0)) {
	top_lines = -1;
#ifdef DEBUG
	fprintf(stderr, "resetting pastheader from value %d\n", pastheader);
#endif
	pastheader=0; /* this is just as dodgey as your "top_lines" ;) */
	getmessage(request, i-1);
  }
  else {
	PSOUT("-ERR No such message");
	PCRLF;
	register_command( NULL, NULL, 0, NULL, NULL );
  }
}

/* 20020923 RJVB: implementation for the top command, as I understood it from a true pop server */
void httpmail_top(cmdtable *command, char *arg, char *arg2)
{
  unsigned long i;
  long lines= strtol(arg2, NULL, 10);

  if (lines < 0) {
	/* For an invalid lines count, just return the whole body */
	httpmail_retrieve(command,arg);
  } else {
	if ((i = strtoul(arg, NULL, 10)) && (i <= inbox_props->listcount) && (i > 0)) {
	  top_lines = lines;
	  cur_message = i-1;
#ifdef DEBUG
	  fprintf(stderr, "resetting pastheader from value %d\n", pastheader);
#endif
	  pastheader=0; 
	  getmessage(request, cur_message);
	} else {
	  PSOUT("-ERR No such message");
	  PCRLF;
	  register_command( NULL, NULL, 0, NULL, NULL );
	}
  }
}

/* This function does the generic work of preparing a http request and sending
 * it. It handles any proxy servers that have been specified on the command
 * line and checks for weird errors as a result of the proxy server being
 * incorrect or unavailable. This will catch some other strange responses that
 * hotmail gives, like when you try to enter an incorrect password for an
 * account many times, perhaps better error reporting will be required here in
 * the future. */
static void prepare_and_send(ghttp_request *request, char *url)
{ 
   ghttp_set_header(request, http_hdr_User_Agent, myAgent); 
   ghttp_set_header(request, http_hdr_Connection, "Keep-Alive"); 
   if (proxy) ghttp_set_proxy(request, proxy);
   if (proxy_username && proxy_password) 
	 ghttp_set_proxy_authinfo(request, proxy_username, proxy_password);   
   my_set_cookies(request, url);
   
   /* Prepare the connection */
   ghttp_prepare(request);

   /* Process the request and exit on error */
   if (ghttp_process(request) == ghttp_error) {
	 snprintf(buffer, N_BUFFER, "-ERR Fatal connection error: %s\015\012", ghttp_get_error(request));
	 PSOUT(buffer);
	 PCRLF;
	 if (log_level > 0) 
	   LOG("Fatal connection error: %s", ghttp_get_error(request));
	 httpmail_destroy();
	 exit(-1);
   }
#ifdef DEBUG
   /* here you can see the status returned along with the "reason" */
   printf("Status returned: %d, reason: %s\n",ghttp_status_code(request), ghttp_reason_phrase(request));
#endif
}
